// exkharc.cpp, v1.0 2014/03/16
// coded by asmodean

// contact: 
//   web:   http://asmodean.reverse.net
//   email: asmodean [at] hush.com
//   irc:   asmodean on efnet (irc.efnet.net)

// This tool extracts ARCC (*.arc) archives used by C̉Ԍt.

#include "as-util.h"

#pragma pack(1)

struct ARCCHDR {
  uint8_t  signature[4]; // "ARCC"
  uint32_t type;
  uint32_t unknown1;
  uint16_t crc1;
  uint16_t unknown2; // 0x1FF1
  uint32_t unknown3;
  uint32_t entry_count;
  uint32_t unknown4;
  uint32_t unknown5;
  uint32_t unknown6;
  uint32_t unknown7;
  uint16_t crc2;
};

struct ARCCTAG_NAME {
  uint8_t  signature[4]; // "NAME"
  uint64_t filename_extent;
  uint16_t crc;
};

struct ARCCTAG_NIDX {
  uint8_t signature[4]; // "NIDX", "EIDX", "CINF", "ADDR"
};

typedef ARCCTAG_NIDX ARCCTAG_EIDX;
typedef ARCCTAG_NIDX ARCCTAG_CINF;
typedef ARCCTAG_NIDX ARCCTAG_ADDR;

struct ARCCTAG_NIDXENTRY {
  uint16_t unknown1;
  uint32_t offset;
  uint16_t crc;
};

typedef ARCCTAG_NIDXENTRY ARCCTAG_EIDXENTRY;

struct ARCCTAG_CINFENTRY {
  uint16_t unknown;
  uint32_t index;
  uint16_t length;
  uint16_t crc;
  
  string filename() const {
    string s((char*) (this + 1), length);

    for (auto& c : s) {
      c ^= 0x69;
    }

    return s;
  }
};

struct ARCCTAG_ADDRENTRY {
  uint16_t unknown1;
  uint64_t offset;
  uint16_t crc;
};

struct ARCCTAG_FILE {
  uint8_t  signature[4]; // "FILE"
  uint32_t unknown1;
  uint32_t unknown2;
  uint32_t unknown3;
  uint32_t unknown4;
  uint32_t unknown5;
  uint32_t length;
  uint8_t  signature2[4]; // "DAEN"
  uint16_t crc;
  // uint8_t data[length];
  // uint16_t crc2;
};

#pragma pack()

int main(int argc, char** argv) {
  if (argc != 2) {
    fprintf(stderr, "exkharc v1.0, coded by asmodean\n\n");
    fprintf(stderr, "usage: %s <input.arc>\n", argv[0]);
    return -1;
  }

  as::file_t input(argv[1]);

  auto hdr           = input.get<ARCCHDR>();
  auto name_tag      = input.get<ARCCTAG_NAME>();
  auto nidx_tag      = input.get<ARCCTAG_NIDX>();
  auto nidx_entries  = input.get_multiple<ARCCTAG_NIDXENTRY>(hdr.entry_count);
  auto eidx_tag      = input.get<ARCCTAG_EIDX>();
  auto eidx_entries  = input.get_multiple<ARCCTAG_EIDXENTRY>(hdr.entry_count);
  auto filename_buff = input.read(name_tag.filename_extent, 0);
  auto addr_tag      = input.get<ARCCTAG_ADDR>();
  auto addr_entries  = input.get_multiple<ARCCTAG_ADDRENTRY>(hdr.entry_count);  

  for (uint32_t i = 0; i < hdr.entry_count; i++) {
    auto&    cinf_entry = filename_buff.get<ARCCTAG_CINFENTRY>(nidx_entries[i].offset);
    auto     file_tag   = input.get<ARCCTAG_FILE>(addr_entries[i].offset);
    auto     buff       = input.read(file_tag.length);

    as::file_t::write(cinf_entry.filename(), buff);
  }

  return 0;
}
