// exkhwag.cpp, v1.0 2014/03/16
// coded by asmodean

// contact: 
//   web:   http://asmodean.reverse.net
//   email: asmodean [at] hush.com
//   irc:   asmodean on efnet (irc.efnet.net)

// This tool extracts IAF_ (*.wag) archives used by C̉Ԍt.

#include "as-util.h"

#pragma pack(1)

struct IAFHDR {
  uint8_t  signature[4]; // "IAF_"
  uint16_t type;
  uint32_t entry_count;
  char     name[64];
  // IAFTOCENTRY entries[entry_count];
  // uint16_t    crc;
};

struct IAFTOCENTRY {
  uint32_t offset;
};

struct IAFTAG {
  uint8_t  signature[4]; // "DATA", "IMGD", "FNNE"
  uint32_t length;
  uint16_t crc;

  IAFTAG() {
    memset(this, 0, sizeof(*this));
  }

  IAFTAG(const char* sig) {
    strncpy((char*)signature, sig, sizeof(signature));
  }

  string get_signature() const {
    return string((char*)signature, sizeof(signature));
  }

  bool operator<(const IAFTAG& rhs) const {
    return get_signature() < rhs.get_signature();
  }
};

struct IMGDHDR {
  uint32_t unknown1; // always zero...
  uint16_t crc;
};

struct IMGDCNTR {
  uint8_t  signature[4];
  uint32_t x;
  uint32_t y;
  uint16_t crc;
};

#pragma pack()

void unobfuscate(as::file_t& file,
                 uint64_t    offset,
                 uint8_t*    buff, 
                 uint32_t    len)
{
  auto end = buff + len;

  while (buff < end) {
    *buff = _rotl8(*buff, 4);
    buff++;
  }
}

int main(int argc, char** argv) {
  if (argc != 2) {
    fprintf(stderr, "exkhwag v1.0, coded by asmodean\n\n");
    fprintf(stderr, "usage: %s <input.wag>\n", argv[0]);
    return -1;
  }

  as::file_t input(argv[1]);

  string prefix = input.get_prefix(true);
  auto   hdr    = input.get<IAFHDR>();

  input.set_filter(unobfuscate);

  auto entries = input.get_multiple<IAFTOCENTRY>(hdr.entry_count);

  uint32_t n = 0;
  for (auto& entry : entries) {
    auto data_tag = input.get<IAFTAG>(entry.offset);

    typedef map<IAFTAG, as::buffer_t> data_t;
    data_t data;

    for (uint32_t i = 0; i < data_tag.length; i++) {
      auto tag = input.get<IAFTAG>();

      if (data.find(tag) != data.end()) {
        fprintf(stderr, "%s: warning: index %d has duplicate tag: %s\n", 
                n,
                string((char*)tag.signature, sizeof(tag.signature)).c_str());
      }

      auto& buff = data[tag] = input.read(tag.length);
    }

    string filename = prefix + as::stringf("+%05d", n++);

    auto fnne = data.find("FNNE");
    if (fnne != data.end()) {
      filename = ".\\" + fnne->second.get_string(0, fnne->second.length() - 2);

      data.erase(fnne);

      as::replace_substr(filename, ":", "");
    }

    auto imgd = data.find("IMGD");
    if (imgd != data.end()) {
      auto& buff      = imgd->second;
      auto& imgdhdr   = buff.get<IMGDHDR>();
      auto  imgdbuff  = buff.offset(sizeof(imgdhdr));
      auto  imgdcntr  = imgdbuff.get<IMGDCNTR>(imgdbuff.length() - sizeof(IMGDCNTR));

      // This is awkward... better way?
      if (!memcmp(imgdcntr.signature, "CNTR", 4)) {
        imgdbuff = imgdbuff.offset(0, imgdbuff.length() - sizeof(imgdcntr));

        if (imgdcntr.x || imgdcntr.y) {
          filename = as::get_file_prefix(filename) 
                       + as::stringf("+x%dy%d", imgdcntr.x, imgdcntr.y) 
                       + as::get_file_extension(filename);
        }
      }

      as::file_t::write(filename, imgdbuff);

      data.erase(imgd);
    }

    uint32_t i = 0;
    for (auto& unknown : data) {
      as::file_t::write(filename + "+" + unknown.first.get_signature(), 
                        unknown.second);
    }
  }

  return 0;
}
